#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/wait.h>

volatile struct shared { double pi; int lock; } *shared;

inline static int xchg(register int reg,
volatile int * volatile obj)
{
  /* Atomic exchange instruction */
__asm__ __volatile__ ("xchgl %1,%0"
                      :"=r" (reg), "=m" (*obj)
                      :"r" (reg), "m" (*obj));
  return(reg);
}

int
main(int argc, char **argv)
{
  register double width, localsum;
  register int intervals, i;
  register int shmid;
  register int iproc = 0;

  /* map an anonymous chunk of memory */
  shared = ((volatile struct shared *)
	    mmap(NULL,				/* put it anywhere */
	         sizeof(struct shared),		/* this big */
		 PROT_READ | PROT_WRITE,	/* read & write */
		 MAP_SHARED | MAP_ANONYMOUS,	/* same seen by all */
		 -1,				/* no file (system swap) */
		 0));
  if (shared == MAP_FAILED) {
    fprintf(stderr, "%s: could not map anonymous segment\n", argv[0]);
    exit(2);
  }

  /* Initialize... */
  shared->pi = 0.0;
  shared->lock = 0;

  /* Fork a child */
  if (!fork()) ++iproc;

  /* get the number of intervals */
  intervals = atoi(argv[1]);
  width = 1.0 / intervals;

  /* do the local computations */
  localsum = 0;
  for (i=iproc; i<intervals; i+=2) {
    register double x = (i + 0.5) * width;
    localsum += 4.0 / (1.0 + x * x);
  }
  localsum *= width;

  /* Atomic spin lock, add, unlock... */
  while (xchg((iproc + 1), &(shared->lock))) ;  
  shared->pi += localsum;
  shared->lock = 0;

  /* Terminate child (barrier sync) */
  if (iproc == 0) {
    wait(NULL);
    printf("Estimation of pi is %f\n", shared->pi);
  }

  /* Check out */
  return(0);
}
